////////////////////////////////////////////////////////////////
// Intro. to Robotic Algorithms
// Assignment 2
//
// You can put all your code in this file.  I've even marked places
// for you to insert your code --- search for "YOUR CODE HERE"
// 
////////////////////////////////////////////////////////////////
#include <svl/SVL.h>
#include <iostream>
#include <list>

#include "world.h"
#include "locproblem.h"
#include <math.h> 

//compensate for windows deficiency
#ifndef M_PI
#define M_PI 3.1415926535897932384626433832
#endif

// global current estimate and covariance
// 
// these need to be general matrices because you will be using some
// non-square matrices in your computations
Mat xhat(3,1);
Mat Px(3,3);

// some globals for use in multiple routines
//
Graphics G;
LocProblem *LP;
objectList locGraphics;

drawableLineStrip path;
drawableSegment beam;
drawablePoint trueLoc, locEstimate;
drawableEllipse locUncertainty;
drawableSegment trueOrient;
drawableSegment orientEstimate;
drawableSegment orientUncertaintyLo, orientUncertaintyHi;

// prototypes
void updateLocalizationGraphics(const Mat& x, const Mat& P, 
				bool addPathPoint);
void initLocalizationGraphics(objectList &L);
void myKeyboardHandler(unsigned char key);
void merge_sensing(const Command& c, double range);
void compound_odometry(const Command& c);
void undisplayBeam();
void displayBeam(const Command& c, double range);


int main(int argc, char *argv[])
{
  // make sure there are enough arguments
  if (argc < 2) {
    cerr << "Usage: " << argv[0] << " [GLUT options] <problem filename>" <<
      endl;
    exit(1);
  }

  // read in the problem (this must be done before graphics initialization)
  LP = new LocProblem(argv[argc - 1]);

  // set initial robot state and covariance
  xhat = Mat(3, 1, LP->startX, LP->startY, LP->startAng);
  Px = Mat(3, 3, vl_zero);			// the zero matrix
  
  // set up graphics objects
  objectList workspace;
  initLocalizationGraphics(locGraphics);
  workspace.push_back(LP->w->boundary);
  copy(LP->w->obstacles.begin(), LP->w->obstacles.end(), 
       back_insert_iterator<objectList>(workspace));
  workspace.push_back(LP->r);
  updateLocalizationGraphics(xhat, Px, true);

  // initialize the graphics --- make the window slightly bigger than
  // the world boundary
  const double marginFrac = 0.075;
  Bbox bounds = LP->w->boundary->bbox();
  double Xmargin = marginFrac * (bounds.xmax() - bounds.xmin());
  double Ymargin = marginFrac * (bounds.ymax() - bounds.ymin());
  G.initGL( &argc, argv, "IRA Assignment 2", 640, 640 );
  G.setBackgroundColor("gray50");
  G.setWorldBounds(Bbox(bounds.xmin()-Xmargin, bounds.ymin()-Ymargin,
			bounds.xmax()+Xmargin, bounds.ymax()+Ymargin));
  G.setKeyboardHandler( myKeyboardHandler );
  G.addObjectList(workspace);
  G.addObjectList(locGraphics);
  G.runEventLoop();

  return 0;
}

void initLocalizationGraphics(objectList &L)
{
  // set default properties
  drawableLineObject::setDefaultLineWidth(2.0f);
  drawableEnclosedObject::setDefaultLineWidth(2.0f);
  drawablePoint::setDefaultPointSize(5.0f);

  // add everything to the objectList
  L.push_back(&path);
  L.push_back(&locUncertainty);
  L.push_back(&trueOrient);
  L.push_back(&trueLoc);
  L.push_back(&orientEstimate);
  L.push_back(&orientUncertaintyLo);
  L.push_back(&orientUncertaintyHi);
  L.push_back(&locEstimate);

  // set colors and other properties
  locUncertainty.setFilled(false);
  locUncertainty.setLineWidth(3.0);
  locUncertainty.setColor(Color("ivory"));
  path.setColor(Color("yellow"));
}


// return a segment that starts at the point "loc" and is at the
// orientation "ang"
// 
//  drawableSegment orientationSegment(Point loc, double ang)
Segment orientationSegment(Point loc, double ang)
{
  const float orientationSegmentLength = 0.15;

//    return drawableSegment(loc, 
  return Segment(loc, 
			 Point(loc.x()+orientationSegmentLength*cos(ang),
			       loc.y()+orientationSegmentLength*sin(ang)));
}

// update the display of localization information using the given
// state estimate "x" and its covariance matrix "P".  It gets the true
// configuration of the robot directly to display this information.
//
// If "addPathPoint" is true, then "x" will be used to add another
// segment to the path.  If false, then "x" will be used to revise the
// last point on the path.
// 
void updateLocalizationGraphics(const Mat& x, const Mat& P, 
				bool addPathPoint)
{
  double tx,ty,ta;
  LP->r->getTruePosition(&tx, &ty, &ta);
  trueLoc = Point(tx,ty);
  trueOrient = orientationSegment(trueLoc, ta);
  trueOrient = orientationSegment(trueLoc, ta);
  const double rx = x[0][0];
  const double ry = x[1][0];
  const double ra = x[2][0];
  LP->r->setConfig(rx, ry, ra);
  locEstimate = Point(rx,ry);
  orientEstimate = orientationSegment(locEstimate, ra);
  if (addPathPoint)
    path.push_back(locEstimate);
  else
    path.back() = locEstimate;
  
  if (P.Rows() != 3 || P.Cols() != 3) {
    cerr << "Hey, your covariance matrix is the wrong size!!!!!" << endl
	 << "ERROR in updateLocalizationGraphics()" << endl;
    exit(1);
  }

  //----------------------------------------------------------------
  // YOUR CODE HERE
  //
  // You need to set the parameters of the ellipse and of the
  // orientation uncertainty.  I've included some random stuff here
  // just so something is drawn for now.
  // 
  double hiAng =  10*M_PI/180.0;
  double loAng = -10*M_PI/180.0;
  orientUncertaintyHi = orientationSegment(locEstimate, hiAng);
  orientUncertaintyLo = orientationSegment(locEstimate, loAng);

  // set the elliptical confidence limits on the position
  //
  // random orientation and fixed radii for an example
  locUncertainty.setPos(locEstimate);
  locUncertainty.setOrientation(rand()/(RAND_MAX+1.0)*2*M_PI);
  locUncertainty.setMajorRadius(0.2);
  locUncertainty.setMinorRadius(0.1);
  //----------------------------------------------------------------


  // set colors (these have to be here because the old colors of
  // drawableSegments and drawablePoints are clobbered when the global
  // is assigned a new drawable object)
  orientEstimate.setColor(Color("purple"));
  orientUncertaintyLo.setColor(Color("magenta"));
  orientUncertaintyHi.setColor(Color("magenta"));
  trueOrient.setColor(Color("lightgreen"));
  trueLoc.setColor(Color("green"));
  locEstimate.setColor(Color("orchid"));
}


void myKeyboardHandler(unsigned char key)
{
  static bool beamDisplayed = false;
  static double sensorReading;
  static Command c;

  switch (key) {
  case 'h': 
    cout << "Commands:" << endl
	 << " n: execute next robot command" << endl
	 << " h: Help" << endl
	 << " q: Quit" << endl;
    G.printControlKeyBindings();
    break;
  case 'n': 
    if (!beamDisplayed) {
      if (LP->c->commandsEnd()) {
	cout << endl << "Done executing all robot commands" << endl;
	break;
      }
      c = LP->c->nextCommand();
      cout << endl << "Executing command: " << c << endl;
      if (c.type == Command::move) {
	compound_odometry(c);	// move the robot and update "xhat" and "Px"
	updateLocalizationGraphics(xhat, Px, true);
      }
      else {			// get the sensor reading and display beam
	beamDisplayed = true;
	sensorReading = LP->r->getRange(c.ang);
	displayBeam(c, sensorReading);
      }
    }
    else {			// remove beam from display...
      undisplayBeam();
      beamDisplayed = false;
      merge_sensing(c, sensorReading); // update "xhat" and "Px"
      updateLocalizationGraphics(xhat, Px, false);
    }
    break;
  case 'q': 
    exit(0);
  }
}

void displayBeam(const Command& c, double range)
{
  // calculate the beam endpoint
  Vec2 robot_pos(xhat[0][0], xhat[1][0]);
  Vec2 pos(range, 0.0);
  Vec2 beamEnd = Rot2(c.ang + xhat[2][0])*pos + robot_pos;

  beam = drawableSegment(Point(xhat[0][0], xhat[1][0]),
			 Point(beamEnd[0], beamEnd[1]));
  beam.setColor(Color("white"));

  locGraphics.push_back(&beam);
}

void undisplayBeam()
{
  locGraphics.remove(&beam);
}


void compound_odometry(const Command& c)
{
  // move the robot
  LP->r->move(c.dist, c.ang);

  //----------------------------------------------------------------
  // YOUR CODE HERE --- update "xhat" and "Px"
  // this will be the preliminary estimates xhat(k+1|k) and Px(k+1|k)
  //
  // I'll just do something simple for now: a random move (that stays
  // inside the world)
  Vec2 rand_move;
  Bbox wb = LP->w->boundary->bbox();
  do {
    double rand_ang = rand()/(RAND_MAX+1.0)*2*M_PI;
    rand_move = Rot2(rand_ang)*Vec2(1.0, 0.0);
  }
  while (rand_move[0]+xhat[0][0] < wb.xmin() || 
	 rand_move[0]+xhat[0][0] > wb.xmax() ||
	 rand_move[1]+xhat[1][0] < wb.ymin() || 
	 rand_move[1]+xhat[1][0] > wb.ymax());
  xhat = xhat + Mat(3, 1, rand_move[0], rand_move[1], 25*M_PI/180.0);
  //----------------------------------------------------------------

  cout << "The robot state estimate is now: x=" << xhat[0][0]
       << ", y=" << xhat[1][0]
       << ", th=" << xhat[2][0]*180/M_PI << endl;
  cout << "The covariance is:" << endl << Px << endl;
}

void merge_sensing(const Command& c, double range)
{
  //----------------------------------------------------------------
  // YOUR CODE HERE --- update "xhat" and "Px"
  // the current values are preliminary estimates xhat(k+1|k) and Px(k+1|k)
  // after this call, they should be the final estimates xhat(k+1) Px(k+1)

  // I'll just do something simple for now...
  xhat = xhat + Mat(3, 1, -0.1, -0.05, -12.5*M_PI/180.0);
  //----------------------------------------------------------------

  cout << "The robot state estimate is now: x=" << xhat[0][0]
       << ", y=" << xhat[1][0]
       << ", th=" << xhat[2][0]*180/M_PI << endl;
  cout << "The covariance is:" << endl << Px << endl;
}
