////////////////////////////////////////////////////////////////
//
// simrobot.h
// Copyright (c) 2002 by Wesley H. Huang.  All rights reserved.
// 
////////////////////////////////////////////////////////////////
#ifndef SIMROBOT_H
#define SIMROBOT_H

#include "world.h"

class SimRobot : public Robot {
private:    
  double trueX, trueY, trueAng;
  World *w;

public:
  double commandDstSigmaRatio;
  double commandAngSigmaRatio;
  double rangeSigmaRatio;

  // (X, Y, Ang) is the initial configuration
  SimRobot(const char *robotFile, double X, double Y, double Ang,
	   World *W, double commandDstSigmaRatio, double commandAngSigmaRatio,
	   double rangeSigmaRatio, bool verbose=false);

  // go forwards for a "dist" and then turn angle "ang" (in radians)
  void move(double dist, double ang);  

  // this is only so that you can plot the point on the screen to see
  // how far off your estimate is.  
  void getTruePosition(double *x, double *y, double *ang) const;

  // This function simulates the laser rangefinder.  It points the
  // sensor in the given direction ("ang" in radians, relative to the
  // robot's forward direction), and measures the distance to the
  // first obstacle along that line.  Gaussian noise is added to the
  // measurement before returning it.
  double getRange(double ang, bool verbose=false);

  // This function is what you use with your Kalman filter.  You have
  // to tell it where you think you are, and it will tell you exactly
  // what the laser rangefinder would return from that configuration
  // without any noise added.  It also gives you the derivatives of H
  // that you need for the Kalman filter.
  double predictRange(double ang, double x, double y, double th,
	       double *dhdx, double *dhdy, double *dhdth,
	       bool verbose=false);

};


////////////////////////////////////////////////////////////////
//
// Implementation of a random number generator that draws from a
// gaussian distribution.  The underlying (uniform deviate)
// pseudorandom number generator is the Park and Miller "Minimum
// standard" generator with Bayes-Durham shuffle
//
// Copyright (c) 2001 Wesley H. Huang.  All rights reserved.
//
////////////////////////////////////////////////////////////////
float gaussian(float mean, float stdev);
void myRandSeed(long s);





#endif
